html {
    height: 100%;
    font-size: 24px;
    font-family: monospace;
}

body {
    height: 100%;
    padding: 0;
    margin: 0;
    display: flex;
}

.left-viewer {
    background-color: rgb(228, 254, 255);
}

.right-viewer {
    background-color: rgb(220, 220, 220);
}

.container {
    display: none;
}
