themes: dict[str, dict[str, str]] = {
    "light": {
        "prompt": "gray",
        "code": "blue",
        "info": "cyan",
        "failure": "dark_red",
        "success": "green",
        "input": "bright_blue",
        "error": "red",
        "warning": "yellow",
    },
    "dark": {
        "prompt": "white",
        "code": "blue",
        "info": "cyan",
        "failure": "bright_red",
        "success": "green",
        "input": "bright_blue",
        "error": "red",
        "warning": "yellow",
    },
}
