{% from 'transcript.jinja' import transcript_container %}
<!DOCTYPE html>
<html>
    <head>
        <link href="https://fonts.googleapis.com/icon?family=Material+Icons" rel="stylesheet">
        <style>
            {% include 'css/conversation_viewer.css' %}
            {% include 'css/transcript.css' %}
        </style>
        <script>
            {% include 'js/conversation_viewer.js' %}
            {% include 'js/transcript.js' %}
        </script>
    </head>
    <body>
        {% for transcript in transcripts %}
            {{ transcript_container(transcript) }}
        {% endfor %}
    </body>
</html>
