html, body {
    height: 100%;
    margin: 0;
    padding: 0;
    font-family: Arial, sans-serif;
}

#header {
    background-color: #f5f5f5;
    padding: 10px;
    display: flex;
    justify-content: space-between;
    align-items: center;
    border-bottom: 1px solid #ddd;
}

#header p {
    margin: 0;
    padding: 0 10px;
    display: inline-block;
}

#container {
    display: flex;
    flex-grow: 1;
    height: calc(100% - 40px); /* Adjust height based on header/footer */
}

#viewer {
    max-width: 85%; /* Test cases for roman numerals involve long words */
}

#selector {
    padding: 10px;
    border-right: 1px solid #ddd;
    overflow-y: auto;
    width: 200px; /* Fixed width for selector */
}

.result-section {
    padding: 10px;
    display: none; /* Initially hidden; will be shown based on the current hash */
    overflow-y: auto;
    flex-grow: 1;
}

button {
    display: block;
    margin-bottom: 5px;
    padding: 10px;
    background-color: #f9f9f9;
    border: 1px solid #ddd;
    cursor: pointer;
    text-align: left;
    width: 100%;
    box-sizing: border-box;
}

button:hover {
    background-color: #e9e9e9;
}

button.green {
    background-color: #d4edda;
    color: #155724;
    border-color: #c3e6cb;
}

button.yellow {
    background-color: #fff3cd;
    color: #856404;
    border-color: #ffeeba;
}

button.grey {
    background-color: #f8f9fa;
    color: #6c757d;
    border-color: #f8f9fa;
}

button.red {
    background-color: #f8d7da;
    color: #721c24;
    border-color: #f5c6cb;
}

.hidden {
    display: none;
}

.content {
    white-space: pre-wrap;
    word-break: break-word;
    background-color: #fff;
    border-bottom: 1px solid #ddd;
    padding: 10px;
    margin-bottom: 10px;
    overflow-y: scroll;
    max-height: 70vh;
}

.family-child {
    margin-left: 20px;
}

.content .container {
    display: flex;
}

h1 {
    margin-top: 0;
}
